<?php

declare(strict_types=1);

namespace mirzaev\vk;

use mirzaev\vk\robots\robot;
use mirzaev\vk\traits\singleton;
use mirzaev\vk\loggers\jasmo;

use Exception;

/**
 * Ядро
 *
 * @property-read int $robots Количество роботов
 * @property string $timezone Временная зона (журналирование)
 * @property array $path Пути (архитектура проекта)
 *
 * @method protected static function __construct() Инициализация
 * @method public static function init() Запуск инициализации или получение инстанции
 * @method public public function build() Сборщик
 * @method public function set($id, $value) Запись в реестр
 * @method public function get($id = null) Чтение из реестра
 *
 * @package mirzaev\vk
 * @author Arsen Mirzaev Tatyano-Muradovich <red@hood.su>
 */
final class core
{
    use singleton;

    /**
     * Счётчик роботов
     *
     * @var int
     */
    private int $robots = 0;

    /**
     * Реестр роботов
     *
     * @var array
     */
    private array $registry = [];

    /**
     * Временная зона
     *
     * Используется в логировании
     *
     * @var string
     */
    private string $timezone;

    /**
     * Путь до корня проекта
     *
     * @var string
     */
    private string $path_root;

    /**
     * Путь до папки журналов
     *
     * @var string
     */
    private string $path_logs;

    /**
     * Путь до временной папки
     *
     * @var string
     */
    private string $path_temp;

    /**
     * Запись в журнал
     *
     * @param string $file Файл для журналирования
     *
     * @todo Добавить установку иного журналиста по спецификации PSR-3
     * @todo Более гибкое журналирование
     */
    public function journal(string $file = null): self
    {
        // Инициализация журналиста (требует переработки)
        jasmo::init()::post($file)::postErrorHandler()::postShutdownHandler();

        return $this;
    }

    /**
     * Записать в реестр
     *
     * @param int $id
     * @param robot $robot
     *
     * @see mirzaev\vk\traits\registry Модификация метода
     */
    public function write(int $id, robot $robot): void
    {
        try {
            // Инициализация уникального идентификатора сессии

            $session = count($this->read($id));
        } catch (Exception $e) {
            if ($e->getCode() === 404) {
                // Робота или сессии не существует

                $session = 0;
            }
        } finally {
            // Записать новую сессию
            $this->registry[$id][$session] = $robot;

            // Прединкрементация счётчика роботов
            ++$this->robots;
        }
    }

    /**
     * Прочитать из реестра
     *
     * Если не передать идентификатор, то вернёт все значения
     *
     * @param int|null $id Идентификатор
     * @param int|null $session Сессия
     *
     * @see mirzaev\vk\traits\registry Модификация метода
     *
     * @return mixed Весь реестр, робота или сессию робота
     */
    public function read(int|null $id = null, int|null $session = null): mixed
    {
        if (isset($id)) {
            // Робот передан

            if (array_key_exists($id, $this->registry)) {
                // Робот найден

                if (isset($session)) {
                    // Сессия робота передана

                    if (array_key_exists($session, $this->registry[$id])) {
                        // Сессия робота найдена

                        return $this->registry[$id][$session];
                    }

                    throw new Exception("Сессия $session робота с идентификатором $id не найдена", 404);
                }

                return $this->registry[$id];
            } else {

                throw new Exception("Робот с идентификатором $id не найден", 404);
            }
        }

        return $this->registry;
    }

    /**
     * Удалить из реестра
     *
     * @param int|null $id Идентификатор
     * @param int|null $session Сессия
     *
     * @see mirzaev\vk\traits\registry Модификация метода
     */
    public function delete(int|null $id = null, int|null $session = null): void
    {
        if (isset($id)) {
            // Робот передан

            if (array_key_exists($id, $this->registry)) {
                // Робот найден

                if (isset($session)) {
                    // Сессия передана

                    if (array_key_exists($session, $this->registry[$id])) {
                        // Сессия найдена

                        // Постдекрементация счётчика роботов
                        --$this->robots;

                        // Удаление сессии
                        unset($this->registry[$id][$session]);

                        return;
                    }

                    throw new Exception("Сессия $session робота с идентификатором $id не найдена", 404);
                }

                // Вычитание из счётчика количества сессий робота
                $this->robots = $this->robots - count($this->registry[$id]);

                // Удаление робота и всех его сессий
                unset($this->registry[$id]);

                return;
            }

            throw new Exception("Робот с идентификатором $id не найден", 404);
        }

        // Полная очистка
        $this->registry = [];
    }

    /**
     * Записать свойство
     *
     * @param mixed $name Название
     * @param mixed $value Значение
     */
    public function __set(mixed $name, mixed $value): void
    {
        match ($name) {
            'timezone' => !isset($this->timezone) ? $this->timezone = $value : throw new Exception('Запрещено переопределять часовой пояс', 500),
            'path_root' => !isset($this->path_root) ? $this->path_root = $value : throw new Exception('Запрещено переопределять корневой каталог', 500),
            'path_logs' => !isset($this->path_logs) ? $this->path_logs = $value : throw new Exception('Запрещено переопределять каталог журналов', 500),
            'path_temp' => !isset($this->path_temp) ? $this->path_temp = $value : throw new Exception('Запрещено переопределять каталог временных файлов', 500)
        };
    }

    /**
     * Прочитать свойство
     *
     * @param mixed $name Название
     */
    public function __get(mixed $name): mixed
    {
        return match ($name) {
            'robots' => $this->robots,
            'timezone' => $this->timezone ?? $this->timezone = 'Europe/Moscow',
            'path_root' => $this->path_root ?? $this->path_root = dirname(__DIR__),
            'path_logs' => $this->path_logs ?? $this->path_logs = $this->__get('path_root') . '/logs',
            'path_temp' => $this->path_temp ?? $this->path_temp = $this->__get('path_root') . '/temp',
            default => null
        };
    }

    /**
     * Вызвать метод
     *
     * Ищет класс описывающий робота,
     * создаёт и возвращает его объект
     *
     * @param string $method Метод
     * @param array $params Параметры
     */
    public function __call(string $method, array $params): robot
    {
        if (class_exists($robot = '\\mirzaev\\vk\\robots\\' . $method)) {
            // Если найден класс реализующий запрошенного робота
            return new $robot(...$params);
        } else {
            throw new Exception("Робот $method не найден", 404);
        }
    }
}
